#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=invalid-name,missing-docstring
import pathlib

from . import common
from ... import base


CUR_DIR = pathlib.Path(__file__).parent


class TestCase(common.TestCase):

    def test_members(self):
        self.assertNotEqual(self.target, base.TDataCollector.target)
        self.assertEqual(self.target, CUR_DIR.name)
        self.assertEqual(
            self.root,
            CUR_DIR.parent.parent / 'res' / self.target / self.kind
        )
        self.assertTrue(self.datasets)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=invalid-name,missing-docstring
import unittest
import pathlib

from . import collector as TT


CUR_DIR = pathlib.Path(__file__).parent


class Collector(TT.TDataCollector):
    # To avoid error because there are no files with '.json' file extension in
    # tests/res/base/basics/20/.
    pattern = '*.*'
    should_exist = ()  # Likewise.


class TestCase(unittest.TestCase, Collector):

    def setUp(self):
        self.init()

    def test_members(self):
        self.assertTrue(self.target)
        self.assertNotEqual(self.target, TT.TDataCollector.target)
        self.assertEqual(self.target, CUR_DIR.name)

        self.assertTrue(self.root is not None)
        self.assertEqual(
            self.root,
            CUR_DIR.parent / 'res' / self.target / self.kind
        )

        self.assertTrue(self.datasets)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""misc global constants, variables, classes and so on.
"""
from .datatypes import (
    IOI_PATH_STR, IOI_PATH_OBJ, IOI_STREAM,
    IOI_TYPES, IOInfo, IOI_KEYS,
    PathOrIOT, PathOrIOInfoT,
    InDataT, InDataExT, PrimitiveT
)
from .errors import (
    UnknownParserTypeError, UnknownProcessorTypeError, UnknownFileTypeError,
    ValidationError
)


GLOB_MARKER = '*'

__all__ = [
    'IOI_PATH_STR', 'IOI_PATH_OBJ', 'IOI_STREAM', 'IOI_TYPES',
    'IOInfo', 'IOI_KEYS',
    'PathOrIOT', 'PathOrIOInfoT',
    'InDataT', 'InDataExT', 'PrimitiveT',
    'UnknownParserTypeError', 'UnknownProcessorTypeError',
    'UnknownFileTypeError', 'ValidationError',
    'GLOB_MARKER',
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=inherit-non-class,too-few-public-methods
"""anyconfig basic data types.
"""
import pathlib
import typing


IOI_PATH_STR: str = 'path'
IOI_PATH_OBJ: str = 'pathlib.Path'
IOI_STREAM: str = 'stream'

IOI_TYPES: typing.FrozenSet[str] = frozenset(
    (IOI_PATH_STR, IOI_PATH_OBJ, IOI_STREAM)
)


class IOInfo(typing.NamedTuple):
    """Equivalent to collections.namedtuple."""
    src: typing.Union[str, pathlib.Path, typing.IO]
    type: str
    path: str
    extension: str


IOI_KEYS: typing.Tuple[str, ...] = IOInfo._fields

PathOrIOT = typing.Union[str, pathlib.Path, typing.IO]
PathOrIOInfoT = typing.Union[PathOrIOT, IOInfo]


InDataT = typing.Mapping[str, typing.Any]

PrimitiveT = typing.Union[None, int, float, bool, str, InDataT]
InDataExT = typing.Union[PrimitiveT, InDataT]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""JSON backends:

- std.json: python standard JSON support library [default]
- simplejson: https://github.com/simplejson/simplejson

Changelog:

.. versionchanged:: 0.9.8

   - Started to split JSON support modules
"""
from . import default
from ..base import ParserClssT


Parser = default.Parser  # To keep backward compatibility.
PARSERS: ParserClssT = [Parser]

try:
    from .simplejson import Parser as SimpleJsonParser
    PARSERS.append(SimpleJsonParser)
except ImportError:
    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Java properties backend:

- Format to support: Java Properties file, e.g.
  http://docs.oracle.com/javase/1.5.0/docs/api/java/util/Properties.html
- Requirements: None (built-in)
- Development Status :: 4 - Beta
- Limitations:

  - Key and value separator of white spaces is not supported
  - Keys contain escaped white spaces is not supported

- Special options: None

Changelog:

.. versionchanged:: 0.7.0

   - Fix handling of empty values, pointed by @ajays20078
   - Fix handling of values contain strings start with '#' or '!' by
     @ajays20078

.. versionadded:: 0.2

   - Added native Java properties parser instead of a plugin utilizes
     pyjavaproperties module.
"""
import os
import re
import warnings

from . import base


_COMMENT_MARKERS = ('#', '!')


def _parseline(line):
    """
    Parse a line of Java properties file.

    :param line:
        A string to parse, must not start with ' ', '#' or '!' (comment)
    :return: A tuple of (key, value), both key and value may be None

    >>> _parseline(" ")
    (None, '')
    >>> _parseline("aaa:")
    ('aaa', '')
    >>> _parseline(" aaa:")
    ('aaa', '')
    >>> _parseline("aaa")
    ('aaa', '')
    >>> _parseline("url = http://localhost")
    ('url', 'http://localhost')
    >>> _parseline("calendar.japanese.type: LocalGregorianCalendar")
    ('calendar.japanese.type', 'LocalGregorianCalendar')
    """
    pair = re.split(r"(?:\s+)?(?:(?<!\\)[=:])", line.strip(), 1)
    key = pair[0].rstrip()

    if len(pair) < 2:
        warnings.warn(f'Invalid line found: {line}', SyntaxWarning)
        return (key or None, '')

    return (key, pair[1].strip())


def _pre_process_line(line, comment_markers=_COMMENT_MARKERS):
    """
    Preprocess a line in properties; strip comments, etc.

    :param line:
        A string not starting w/ any white spaces and ending w/ line breaks.
        It may be empty. see also: :func:`load`.
    :param comment_markers: Comment markers, e.g. '#' (hash)

    >>> _pre_process_line('') is None
    True
    >>> s0 = "calendar.japanese.type: LocalGregorianCalendar"
    >>> _pre_process_line("# " + s0) is None
    True
    >>> _pre_process_line("! " + s0) is None
    True
    >>> _pre_process_line(s0 + "# comment")
    'calendar.japanese.type: LocalGregorianCalendar# comment'
    """
    if not line:
        return None

    if any(c in line for c in comment_markers):
        if line.startswith(comment_markers):
            return None

    return line


def unescape(in_s):
    """
    :param in_s: Input string
    """
    return re.sub(r"\\(.)", r"\1", in_s)


def _escape_char(in_c):
    """
    Escape some special characters in java .properties files.

    :param in_c: Input character

    >>> "\\:" == _escape_char(':')
    True
    >>> "\\=" == _escape_char('=')
    True
    >>> _escape_char('a')
    'a'
    """
    return '\\' + in_c if in_c in (':', '=', '\\') else in_c


def escape(in_s):
    """
    :param in_s: Input string
    """
    return ''.join(_escape_char(c) for c in in_s)


def load(stream, container=dict, comment_markers=_COMMENT_MARKERS):
    """
    Load and parse Java properties file given as a fiel or file-like object
    'stream'.

    :param stream: A file or file like object of Java properties files
    :param container:
        Factory function to create a dict-like object to store properties
    :param comment_markers: Comment markers, e.g. '#' (hash)
    :return: Dict-like object holding properties

    >>> import io
    >>> to_strm = io.StringIO
    >>> s0 = "calendar.japanese.type: LocalGregorianCalendar"
    >>> load(to_strm(''))
    {}
    >>> load(to_strm("# " + s0))
    {}
    >>> load(to_strm("! " + s0))
    {}
    >>> load(to_strm("calendar.japanese.type:"))
    {'calendar.japanese.type': ''}
    >>> load(to_strm(s0))
    {'calendar.japanese.type': 'LocalGregorianCalendar'}
    >>> load(to_strm(s0 + "# ..."))
    {'calendar.japanese.type': 'LocalGregorianCalendar# ...'}
    >>> s1 = r"key=a\\:b"
    >>> load(to_strm(s1))
    {'key': 'a:b'}
    >>> s2 = '''application/postscript: \\
    ...         x=Postscript File;y=.eps,.ps
    ... '''
    >>> load(to_strm(s2))
    {'application/postscript': 'x=Postscript File;y=.eps,.ps'}
    """
    ret = container()
    prev = ""

    for line in stream:
        line = _pre_process_line(prev + line.strip().rstrip(),
                                 comment_markers)
        # I don't think later case may happen but just in case.
        if line is None or not line:
            continue

        prev = ""  # re-initialize for later use.

        if line.endswith("\\"):
            prev += line.rstrip(" \\")
            continue

        (key, val) = _parseline(line)
        if key is None:
            warnings.warn(f'Failed to parse the line: {line}')
            continue

        ret[key] = unescape(val)

    return ret


class Parser(base.StreamParser):
    """
    Parser for Java properties files.
    """
    _cid = 'properties'
    _type = 'properties'
    _extensions = ['properties']
    _ordered = True
    _dict_opts = ['ac_dict']

    def load_from_stream(self, stream, container, **kwargs):
        """
        Load config from given file like object 'stream'.

        :param stream: A file or file like object of Java properties files
        :param container: callble to make a container object
        :param kwargs: optional keyword parameters (ignored)

        :return: Dict-like object holding config parameters
        """
        return load(stream, container=container)

    def dump_to_stream(self, cnf, stream, **kwargs):
        """
        Dump config 'cnf' to a file or file-like object 'stream'.

        :param cnf: Java properties config data to dump
        :param stream: Java properties file or file like object
        :param kwargs: backend-specific optional keyword parameters :: dict
        """
        for key, val in cnf.items():
            stream.write(f'{key} = {escape(val)}{os.linesep}')

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Utility funtions to process file and file paths.
"""
import pathlib
import typing


def is_io_stream(obj: typing.Any) -> bool:
    """Is given object ``obj`` an IO stream, file or file-like object?
    """
    return callable(getattr(obj, 'read', False))


def get_path_from_stream(strm: typing.IO, safe: bool = False) -> str:
    """
    Try to get file path from given file or file-like object 'strm'.

    :param strm: A file or file-like object might have its file path info
    :return: file path or None
    :raises: ValueError
    """
    if not is_io_stream(strm) and not safe:
        raise ValueError(f'It does not look a file[-like] object: {strm!r}')

    path = getattr(strm, 'name', None)
    if path is not None:
        try:
            return str(pathlib.Path(path).resolve())
        except (TypeError, ValueError):
            pass

    return ''

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import pathlib
import tempfile
import unittest

import anyconfig.schema

from anyconfig.api import ValidationError

from . import common


SCM_NG_0 = '{"type": "object", "properties": {"a": {"type": "string"}}}'


@unittest.skipIf(not anyconfig.schema.SUPPORTED,
                 'jsonschema lib is not available')
class TestCase(common.TestCase):
    kind = 'schema'

    def test_multi_load(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn(
                    tdata.inputs, ac_schema=tdata.scm, **tdata.opts
                ),
                tdata.exp,
                tdata
            )

    def test_multi_load_with_schema_validation_failure(self):
        with tempfile.TemporaryDirectory() as tdir:
            wdir = pathlib.Path(tdir)
            scm = wdir / 'scm.json'
            scm.write_text(SCM_NG_0)

            for tdata in self.each_data():
                with self.assertRaises(ValidationError):
                    self.target_fn(
                        tdata.inputs, ac_schema=scm, ac_schema_safe=False
                    )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
"""File based test data collector.
"""
import ast
import collections
import importlib.abc
import importlib.util
import json
import pathlib
import typing
import warnings

from .datatypes import (
    DictT, MaybePathT, TDataPaths
)


def target_by_parent(self: str = __file__):
    """
    >>> target_by_parent()
    'base'
    """
    return pathlib.Path(self).parent.name


def load_from_py(py_path: typing.Union[str, pathlib.Path],
                 data_name: str = 'DATA') -> DictT:
    """Load from .py files.

    .. note:: It's not safe always.
    """
    spec = importlib.util.spec_from_file_location('testmod', py_path)
    if spec and isinstance(spec.loader, importlib.abc.Loader):
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        try:
            return getattr(mod, data_name, None)
        except (TypeError, ValueError, AttributeError):
            pass

    return {}


def load_literal_data_from_py(py_path: typing.Union[str, pathlib.Path]
                              ) -> DictT:
    """.. note:: It might be safer than the above function.
    """
    return ast.literal_eval(pathlib.Path(py_path).read_text().strip())


def maybe_data_path(datadir: pathlib.Path, name: str,
                    should_exist: typing.Iterable[str] = (),
                    file_ext: str = '*'
                    ) -> typing.Optional[pathlib.Path]:
    """
    Get and return the file path of extra data file. Its filename will be
    computed from the filename of the base data file given.
    """
    pattern = f'{name}.{file_ext}'
    if datadir.exists() and datadir.is_dir():
        paths = sorted(datadir.glob(pattern))
        if paths:
            return paths[0]  # There should be just one file found.

    if datadir.name in should_exist:
        raise OSError(f'{datadir!s}/{pattern} should exists but not')

    return None


def load_data(path: MaybePathT,
              default: typing.Optional[typing.Any] = None,
              should_exist: bool = False,
              exec_py: bool = False,
              ordered: bool = False
              ) -> typing.Union[DictT, str]:
    """
    Return data loaded from given path or the default value.
    """
    if path is None or not path:
        if not should_exist:
            return {} if default is None else default

    if path and path.exists():
        if path.suffix == '.json':
            if ordered:
                return json.load(
                    path.open(), object_hook=collections.OrderedDict
                )

            return json.load(path.open())

        if path.suffix == '.py':
            return (
                load_from_py if exec_py else load_literal_data_from_py
            )(path)  # type: ignore

        if path.suffix == '.txt':
            return path.read_text()

        return path

    raise ValueError(f'Not exist or an invalid data: {path!s}')


def each_data_from_dir(datadir: pathlib.Path, pattern: str = '*.json',
                       should_exist: typing.Iterable[str] = ()
                       ) -> typing.Iterator[TDataPaths]:
    """
    Yield a collection of paths of data files under given dir.
    """
    if not datadir.is_dir():
        raise ValueError(f'Not look a data dir: {datadir!s}')

    for inp in sorted(datadir.glob(pattern)):
        if not inp.exists():
            warnings.warn(f'Not exists: {inp!s}')
            continue

        if not inp.is_file():
            warnings.warn(f'Not looks a file: {inp!s}')
            continue

        name = inp.stem

        yield TDataPaths(
            datadir,
            inp,
            maybe_data_path(datadir / 'e', name, should_exist),
            maybe_data_path(datadir / 'o', name, should_exist),
            maybe_data_path(datadir / 's', name, should_exist),
            maybe_data_path(datadir / 'q', name, should_exist),
            maybe_data_path(datadir / 'c', name, should_exist)
        )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest
import warnings

import anyconfig.api._load as TT
import anyconfig.schema

from anyconfig.api import ValidationError

from . import common


SCM_NG_0 = '{"type": "object", "properties": {"a": {"type": "string"}}}'


@unittest.skipIf(not anyconfig.schema.SUPPORTED,
                 'jsonschema lib is not available')
class TestCase(common.TestCase):
    kind = 'schema'

    def test_loads_with_schema_validation(self):
        for data in self.each_data():
            scm = data.scm.read_text().strip()
            self.assertEqual(
                TT.loads(data.inp, ac_schema=scm, **data.opts),
                data.exp,
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_loads_with_schema_validation_failures(self):
        opts = dict(ac_parser='json', ac_schema=SCM_NG_0)

        for data in self.each_data():
            with warnings.catch_warnings(record=True) as warns:
                warnings.simplefilter('always')
                self.assertTrue(
                    TT.loads(data.inp, **opts) is None,
                    f'{data.datadir!s}, {data.inp_path!s}'
                )
                self.assertTrue(len(warns) > 0)
                self.assertTrue(issubclass(warns[-1].category, UserWarning))

            with self.assertRaises(ValidationError):
                TT.loads(data.inp, ac_schema_safe=False, **opts)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-argument
r"""Dummy implementation of anyconfig.query.*.
"""
from ..common import InDataExT
from .datatypes import MaybeJexp


def try_query(data: InDataExT, jexp: MaybeJexp = None, **options) -> InDataExT:
    """
    Dummy implementation of :func:`anyconfig.query.try_query`
    """
    return data

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-import,import-error,invalid-name
r"""Public APIs of anyconfig module to load configuration files.
"""
import typing
import warnings

from .. import ioinfo
from ..common import (
    InDataT, InDataExT, PathOrIOInfoT
)
from ..dicts import (
    convert_to as dicts_convert_to,
    merge as dicts_merge
)
from ..parsers import find as parsers_find
from ..query import try_query
from ..schema import is_valid
from ..template import try_render
from ..utils import (
    is_dict_like, is_path_like_object, is_paths
)
from .datatypes import (
    ParserT
)
from .utils import are_same_file_types


MappingT = typing.Dict[str, typing.Any]
MaybeParserOrIdOrTypeT = typing.Optional[typing.Union[str, ParserT]]


def _maybe_schema(**options) -> typing.Optional[InDataT]:
    """
    :param options: Optional keyword arguments such as

        - ac_template: Assume configuration file may be a template file and try
          to compile it AAR if True
        - ac_context: Mapping object presents context to instantiate template
        - ac_schema: JSON schema file path to validate configuration files

    :return: Mapping object or None means some errors
    """
    ac_schema = options.get("ac_schema", None)
    if ac_schema is not None:
        # Try to detect the appropriate parser to load the schema data as it
        # may be different from the original config file's format, perhaps.
        options["ac_parser"] = None
        options["ac_schema"] = None  # Avoid infinite loop.
        return load(ac_schema, **options)

    return None


def _single_load(input_: PathOrIOInfoT,
                 ac_parser: MaybeParserOrIdOrTypeT = None,
                 ac_template: bool = False,
                 ac_context: typing.Optional[MappingT] = None,
                 **options) -> InDataExT:
    """
    :param input_:
        File path or file or file-like object or pathlib.Path object represents
        the file or a namedtuple 'anyconfig.common.IOInfo' object represents
        some input to load some data from
    :param ac_parser: Forced parser type or parser object itself
    :param ac_template:
        Assume configuration file may be a template file and try to compile it
        AAR if True
    :param ac_context: A dict presents context to instantiate template
    :param options:
        Optional keyword arguments :func:`single_load` supports except for
        ac_schema and ac_query

    :return: Mapping object
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    ioi = ioinfo.make(input_)
    psr: ParserT = parsers_find(ioi, forced_type=ac_parser)
    filepath = ioi.path

    if ac_template and filepath:
        content = try_render(filepath=filepath, ctx=ac_context, **options)
        if content is not None:
            return psr.loads(content, **options)

    return psr.load(ioi, **options)


def single_load(input_: PathOrIOInfoT,
                ac_parser: MaybeParserOrIdOrTypeT = None,
                ac_template: bool = False,
                ac_context: typing.Optional[MappingT] = None,
                **options) -> InDataExT:
    r"""
    Load single configuration file.

    .. note::

       :func:`load` is a preferable alternative and this API should be used
       only if there is a need to emphasize given input 'input\_' is single
       one.

    :param input\_:
        File path or file or file-like object or pathlib.Path object represents
        the file or a namedtuple 'anyconfig.common.IOInfo' object represents
        some input to load some data from
    :param ac_parser: Forced parser type or parser object itself
    :param ac_template:
        Assume configuration file may be a template file and try to compile it
        AAR if True
    :param ac_context: A dict presents context to instantiate template
    :param options: Optional keyword arguments such as:

        - Options common in :func:`single_load`, :func:`multi_load`,
          :func:`load` and :func:`loads`:

          - ac_dict: callable (function or class) to make mapping objects from
            loaded data if the selected backend can customize that such as JSON
            which supports that with 'object_pairs_hook' option, or None. If
            this option was not given or None, dict or
            :class:`collections.OrderedDict` will be used to make result as
            mapping object depends on if ac_ordered (see below) is True and
            selected backend can keep the order of items loaded. See also
            :meth:`_container_factory` of
            :class:`anyconfig.backend.base.Parser` for more implementation
            details.

          - ac_ordered: True if you want to keep resuls ordered. Please note
            that order of items may be lost depends on the selected backend.

          - ac_schema: JSON schema file path to validate given config file
          - ac_query: JMESPath expression to query data

        - Common backend options:

          - ac_ignore_missing:
            Ignore and just return empty result if given file 'input\_' does
            not exist actually.

        - Backend specific options such as {"indent": 2} for JSON backend

    :return: Mapping object
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    cnf = _single_load(input_, ac_parser=ac_parser, ac_template=ac_template,
                       ac_context=ac_context, **options)
    schema = _maybe_schema(ac_template=ac_template, ac_context=ac_context,
                           **options)
    if schema and not is_valid(cnf, schema, **options):
        return None

    return try_query(cnf, options.get('ac_query', False), **options)


def multi_load(inputs: typing.Union[typing.Iterable[PathOrIOInfoT],
                                    PathOrIOInfoT],
               ac_parser: MaybeParserOrIdOrTypeT = None,
               ac_template: bool = False,
               ac_context: typing.Optional[MappingT] = None,
               **options) -> InDataExT:
    r"""
    Load multiple config files.

    .. note::

       :func:`load` is a preferable alternative and this API should be used
       only if there is a need to emphasize given inputs are multiple ones.

    The first argument 'inputs' may be a list of a file paths or a glob pattern
    specifying them or a pathlib.Path object represents file[s] or a namedtuple
    'anyconfig.common.IOInfo' object represents some inputs to load some data
    from.

    About glob patterns, for example, is, if a.yml, b.yml and c.yml are in the
    dir /etc/foo/conf.d/, the followings give same results::

      multi_load(["/etc/foo/conf.d/a.yml", "/etc/foo/conf.d/b.yml",
                  "/etc/foo/conf.d/c.yml", ])

      multi_load("/etc/foo/conf.d/*.yml")

    :param inputs:
        A list of file path or a glob pattern such as r'/a/b/\*.json'to list of
        files, file or file-like object or pathlib.Path object represents the
        file or a namedtuple 'anyconfig.common.IOInfo' object represents some
        inputs to load some data from
    :param ac_parser: Forced parser type or parser object
    :param ac_template: Assume configuration file may be a template file and
        try to compile it AAR if True
    :param ac_context: Mapping object presents context to instantiate template
    :param options: Optional keyword arguments:

        - ac_dict, ac_ordered, ac_schema and ac_query are the options common in
          :func:`single_load`, :func:`multi_load`, :func:`load`: and
          :func:`loads`. See the descriptions of them in :func:`single_load`.

        - Options specific to this function and :func:`load`:

          - ac_merge (merge): Specify strategy of how to merge results loaded
            from multiple configuration files. See the doc of
            :mod:`dicts` for more details of strategies. The default
            is dicts.MS_DICTS.

        - Common backend options:

          - ignore_missing: Ignore and just return empty result if given file
            'path' does not exist.

        - Backend specific options such as {"indent": 2} for JSON backend

    :return: Mapping object or any query result might be primitive objects
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    schema = _maybe_schema(ac_template=ac_template, ac_context=ac_context,
                           **options)
    options['ac_schema'] = None  # Avoid to load schema more than twice.

    paths = ioinfo.expand_paths(inputs)
    if are_same_file_types(paths):
        ac_parser = parsers_find(paths[0], forced_type=ac_parser)

    cnf = None
    ctx = dicts_convert_to({}, **options)
    if ac_context:
        ctx = ac_context.copy()

    for path in paths:
        cups = _single_load(
            path, ac_parser=ac_parser, ac_template=ac_template,
            ac_context=ctx, **options
        )
        if cups:
            if cnf is None:
                cnf = cups  # type: ignore

            if is_dict_like(cups):
                dicts_merge(
                    typing.cast(MappingT, cnf),
                    typing.cast(MappingT, cups),
                    **options
                )
                dicts_merge(ctx, typing.cast(MappingT, cups), **options)
            elif len(paths) > 1:
                raise ValueError(
                    f'Object loaded from {path!r} is not a mapping object and '
                    'cannot be merged with later ones will be loaded from '
                    'other inputs.'
                )

    if cnf is None:
        return dicts_convert_to({}, **options)

    if schema and not is_valid(cnf, schema, **options):
        return None

    return try_query(cnf, options.get('ac_query', False), **options)


def load(path_specs, ac_parser=None, ac_dict=None, ac_template=False,
         ac_context=None, **options):
    r"""
    Load single or multiple config files or multiple config files specified in
    given paths pattern or pathlib.Path object represents config files or a
    namedtuple 'anyconfig.common.IOInfo' object represents some inputs.

    :param path_specs:
        A list of file path or a glob pattern such as r'/a/b/\*.json'to list of
        files, file or file-like object or pathlib.Path object represents the
        file or a namedtuple 'anyconfig.common.IOInfo' object represents some
        inputs to load some data from.
    :param ac_parser: Forced parser type or parser object
    :param ac_dict:
        callable (function or class) to make mapping object will be returned as
        a result or None. If not given or ac_dict is None, default mapping
        object used to store resutls is dict or
        :class:`collections.OrderedDict` if ac_ordered is True and selected
        backend can keep the order of items in mapping objects.

    :param ac_template: Assume configuration file may be a template file and
        try to compile it AAR if True
    :param ac_context: A dict presents context to instantiate template
    :param options:
        Optional keyword arguments. See also the description of 'options' in
        :func:`single_load` and :func:`multi_load`

    :return: Mapping object or any query result might be primitive objects
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    if is_path_like_object(path_specs):
        return single_load(path_specs, ac_parser=ac_parser, ac_dict=ac_dict,
                           ac_template=ac_template, ac_context=ac_context,
                           **options)

    if not is_paths(path_specs):
        raise ValueError(f'Possible invalid input {path_specs!r}')

    return multi_load(path_specs, ac_parser=ac_parser, ac_dict=ac_dict,
                      ac_template=ac_template, ac_context=ac_context,
                      **options)


def loads(content, ac_parser=None, ac_dict=None, ac_template=False,
          ac_context=None, **options):
    """
    :param content: Configuration file's content (a string)
    :param ac_parser: Forced parser type or ID or parser object
    :param ac_dict:
        callable (function or class) to make mapping object will be returned as
        a result or None. If not given or ac_dict is None, default mapping
        object used to store resutls is dict or
        :class:`collections.OrderedDict` if ac_ordered is True and selected
        backend can keep the order of items in mapping objects.
    :param ac_template: Assume configuration file may be a template file and
        try to compile it AAR if True
    :param ac_context: Context dict to instantiate template
    :param options:
        Optional keyword arguments. See also the description of 'options' in
        :func:`single_load` function.

    :return: Mapping object or any query result might be primitive objects
    :raises: ValueError, UnknownProcessorTypeError
    """
    if ac_parser is None:
        warnings.warn("ac_parser was not given but it's must to find correct "
                      "parser to load configurations from string.")
        return None

    psr = parsers_find(None, forced_type=ac_parser)
    schema = None
    ac_schema = options.get('ac_schema', None)
    if ac_schema is not None:
        options['ac_schema'] = None
        schema = loads(ac_schema, ac_parser=psr, ac_dict=ac_dict,
                       ac_template=ac_template, ac_context=ac_context,
                       **options)

    if ac_template:
        compiled = try_render(content=content, ctx=ac_context, **options)
        if compiled is not None:
            content = compiled

    cnf = psr.loads(content, ac_dict=ac_dict, **options)
    if not is_valid(cnf, schema, **options):
        return None

    return try_query(cnf, options.get('ac_query', False), **options)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
r"""test cases for anyconfig.utils.lists.
"""
import operator
import unittest

import anyconfig.utils.lists as TT


class TestCase(unittest.TestCase):

    def test_groupby(self):
        items = (('a', 1), ('b', -1), ('c', 1))
        res = TT.groupby(items, operator.itemgetter(1))
        self.assertEqual(
            [(key, tuple(grp)) for key, grp in res],
            [(-1, (('b', -1),)), (1, (('a', 1), ('c', 1)))]
        )

    def test_concat(self):
        ies = (
            ([[]], []),
            ((()), []),
            ([[1, 2, 3], [4, 5]], [1, 2, 3, 4, 5]),
            ([[1, 2, 3], [4, 5, [6, 7]]], [1, 2, 3, 4, 5, [6, 7]]),
            (((1, 2, 3), (4, 5, (6, 7))), [1, 2, 3, 4, 5, (6, 7)]),
            (((i, i * 2) for i in range(3)), [0, 0, 1, 2, 2, 4]),
        )
        for inp, exp in ies:
            self.assertEqual(TT.concat(inp), exp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
"""Collector to collect file based test data.
"""
import typing

from ... import base
from . import datatypes, utils


class DataCollector(base.TDataCollector):
    """Data collector for api.multi_load
    """
    def load_datasets(self) -> typing.List[datatypes.TData]:
        """Load test data from files.
        """
        _datasets = sorted(
            utils.each_data_from_dir(
                self.root, self.pattern, self.should_exist
            )
        )
        if not _datasets:
            raise ValueError(f'No data: {self.root!s}')

        for tdata in _datasets:
            if not tdata.inputs:
                raise ValueError(f'No data in subdir: {tdata.subdir!s}')

        return _datasets

    def each_data(self) -> typing.Iterator[datatypes.TData]:
        """Yields test data.
        """
        if not self.initialized:
            self.init()

        for tdata in self.datasets:
            yield tdata

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

from ... import base


class TestCase(unittest.TestCase, base.TDataCollector):
    pattern = '*.txt'

    def setUp(self):
        self.init()

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Abstract processor module.

.. versionadded:: 0.9.5

   - Add to abstract processors such like Parsers (loaders and dumpers).
"""
import typing


class Processor:
    """
    Abstract processor class to provide basic implementation of some methods,
    interfaces and members.

    - _type: type indicates data types it can process
    - _priority: Priority to select it if there are others of same type
    - _extensions: File extensions of data type it can process

    .. note::
       This class ifself is not a singleton but its children classes should so
       in most cases, I think.
    """
    _cid: str = ''
    _type: str = ''
    _priority: int = 0   # 0 (lowest priority) .. 99  (highest priority)
    _extensions: typing.List[str] = []

    @classmethod
    def cid(cls) -> str:
        """Processor class ID.
        """
        return cls._cid

    @classmethod
    def type(cls) -> str:
        """Processors' type.
        """
        return str(cls._type)

    @classmethod
    def priority(cls) -> int:
        """Processors's priority.
        """
        return cls._priority

    @classmethod
    def extensions(cls) -> typing.List[str]:
        """A list of file extensions of files which this process can process.
        """
        return cls._extensions

    @classmethod
    def __eq__(cls, other) -> bool:
        """Test equality.
        """
        return cls.cid() == other.cid()

    def __str__(self) -> str:
        """A string repr.
        """
        return (
            f'<Processor cid={self.cid()}, type={self.type()}, '
            f'prio={self.priority()}, extensions={self.extensions()!r}'
        )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2016 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods
# pylint: disable=ungrouped-imports

from collections import OrderedDict

import anyconfig.backend.shellvars as TT
import tests.backend.common as TBC


CNF = OrderedDict((("a", "0"), ("b", "bbb"), ("c", "ccc"), ("d", "ddd"),
                   ("e", "eee")))


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf = CNF
    cnf_s = TBC.read_from_res("20-00-cnf.sh")


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):

    pass


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):

    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# type() is used to exactly match check instead of isinstance here.
# pylint: disable=unidiomatic-typecheck
r"""YAML backend:

- Format to support: YAML, http://yaml.org
- Requirements: PyYAML (yaml), http://pyyaml.org
- Development Status :: 5 - Production/Stable
- Limitations:

  - Resuls is not ordered even if 'ac_ordered' or 'ac_dict' was given.

- Special options:

  - All keyword options of yaml.safe_load, yaml.load, yaml.safe_dump and
    yaml.dump should work.

  - Use 'ac_safe' boolean keyword option if you prefer to call yaml.safe_load
    and yaml.safe_dump instead of yaml.load and yaml.dump. Please note that
    this option conflicts with 'ac_dict' option and these options cannot be
    used at the same time.

  - See also: http://pyyaml.org/wiki/PyYAMLDocumentation

Changelog:

.. versionchanged:: 0.9.6

   - Add support of loading primitives other than mapping objects.

.. versionchanged:: 0.3

   - Changed special keyword option 'ac_safe' from 'safe' to avoid
     possibility of option conflicts in the future.
"""
import yaml
try:
    from yaml import CSafeLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import SafeLoader as Loader, Dumper  # type: ignore

from ...dicts import convert_to
from ...utils import is_dict_like
from .. import base
from . import common


_MAPPING_TAG = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG


def _customized_loader(container, loader=Loader, mapping_tag=_MAPPING_TAG):
    """
    Create or update loader with making given callble 'container' to make
    mapping objects such as dict and OrderedDict, used to construct python
    object from yaml mapping node internally.

    :param container: Set container used internally
    """
    def construct_mapping(loader, node, deep=False):
        """Construct python object from yaml mapping node, based on
        :meth:`yaml.BaseConstructor.construct_mapping` in PyYAML (MIT).
        """
        loader.flatten_mapping(node)
        if not isinstance(node, yaml.MappingNode):
            raise yaml.constructor.ConstructorError(
                None, None, f'expected a mapping node, but found {node.id}',
                node.start_mark
            )
        mapping = container()
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                eargs = ('while constructing a mapping',
                         node.start_mark,
                         f'found unacceptable key ({exc!s})',
                         key_node.start_mark)
                raise yaml.constructor.ConstructorError(*eargs)
            value = loader.construct_object(value_node, deep=deep)
            mapping[key] = value

        return mapping

    tag = 'tag:yaml.org,2002:python/unicode'

    def construct_ustr(loader, node):
        """Unicode string constructor"""
        return loader.construct_scalar(node)

    try:
        loader.add_constructor(tag, construct_ustr)
    except NameError:
        pass

    if type(container) != dict:
        loader.add_constructor(mapping_tag, construct_mapping)
    return loader


def _customized_dumper(container, dumper=Dumper):
    """
    Counterpart of :func:`_customized_loader` for dumpers.
    """
    def container_representer(dumper, data, mapping_tag=_MAPPING_TAG):
        """Container representer.
        """
        return dumper.represent_mapping(mapping_tag, data.items())

    def ustr_representer(dumper, data):
        """Unicode string representer"""
        tag = 'tag:yaml.org,2002:python/unicode'
        return dumper.represent_scalar(tag, data)

    try:
        dumper.add_representer(unicode, ustr_representer)
    except NameError:
        pass

    if type(container) != dict:
        dumper.add_representer(container, container_representer)
    return dumper


def yml_fnc_by_name(fname, **options):
    """Utility functo to get yaml loading/dumping function by name.

    :param fname:
        "load" or "dump", not checked but it should be OK.
        see also :func:`yml_load` and :func:`yml_dump`
    :param options: keyword args may contain "ac_safe" to load/dump safely
    """
    return getattr(yaml, f'safe_{fname}' if options.get('ac_safe') else fname)


def yml_fnc_(fname, *args, **options):
    """An wrapper of yaml.safe_load, yaml.load, yaml.safe_dump and yaml.dump.

    :param fname:
        "load" or "dump", not checked but it should be OK.
        see also :func:`yml_load` and :func:`yml_dump`
    :param args: [stream] for load or [cnf, stream] for dump
    :param options: keyword args may contain "ac_safe" to load/dump safely
    """
    fnc = yml_fnc_by_name(fname, **options)
    return fnc(*args, **common.filter_from_options('ac_safe', options))


def yml_load(stream, container, yml_fnc=yml_fnc_, **options):
    """An wrapper of yaml.safe_load and yaml.load.

    :param stream: a file or file-like object to load YAML content
    :param container: callble to make a container object

    :return: Mapping object
    """
    if options.get('ac_safe', False):
        # .. note:: yaml.safe_load does not support any keyword options.
        options = dict(ac_safe=True)

    elif not options.get('Loader', False):
        maybe_container = options.get('ac_dict', False)
        if maybe_container and callable(maybe_container):
            container = maybe_container

        options['Loader'] = _customized_loader(container)

    ret = yml_fnc('load', stream,
                  **common.filter_from_options('ac_dict', options))
    if ret is None:
        return container()

    return ret


def yml_dump(data, stream, yml_fnc=yml_fnc_, **options):
    """An wrapper of yaml.safe_dump and yaml.dump.

    :param data: Some data to dump
    :param stream: a file or file-like object to dump YAML data
    """
    _is_dict = is_dict_like(data)

    if options.get('ac_safe', False):
        options = dict(ac_safe=True)  # Same as yml_load.

    elif not options.get('Dumper', False) and _is_dict:
        # TODO: Any other way to get its constructor?
        maybe_container = options.get('ac_dict', type(data))
        options['Dumper'] = _customized_dumper(maybe_container)

    if _is_dict:
        # Type information and the order of items are lost on dump currently.
        data = convert_to(data, ac_dict=dict)
        options = common.filter_from_options('ac_dict', options)

    return yml_fnc('dump', data, stream, **options)


class Parser(common.Parser):
    """
    Parser for YAML files.
    """
    _cid = 'pyyaml'
    _priority = 30  # Higher priority than ruamel.yaml.
    _load_opts = ['Loader', 'ac_safe', 'ac_dict']
    _dump_opts = ['stream', 'ac_safe', 'Dumper', 'default_style',
                  'default_flow_style', 'canonical', 'indent', 'width',
                  'allow_unicode', 'line_break', 'encoding', 'explicit_start',
                  'explicit_end', 'version', 'tags']

    load_from_stream = base.to_method(yml_load)
    dump_to_stream = base.to_method(yml_dump)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, invalid-name
import unittest

import anyconfig.backend.base.loaders as TT


FILE_PATH = __file__


class LoaderMixinTestCase(unittest.TestCase):

    def test_ropen(self):
        with TT.LoaderMixin().ropen(FILE_PATH) as fio:
            self.assertEqual(fio.mode, 'r')


class BinaryLoaderMixinTestCase(unittest.TestCase):

    def test_ropen(self):
        with TT.BinaryLoaderMixin().ropen(FILE_PATH) as fio:
            self.assertEqual(fio.mode, 'rb')

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""Test cases for anyconfig.utils.files.
"""
import pathlib
import tempfile
import unittest

import anyconfig.ioinfo.paths as TT

from anyconfig.ioinfo import make as ioinfo_make


class TestCase(unittest.TestCase):

    def test_expand_paths(self):
        with tempfile.TemporaryDirectory() as workdir:
            tdir = pathlib.Path(str(workdir)) / 'a' / 'b' / 'c'
            tdir.mkdir(parents=True)

            pathlib.Path(tdir / 'd.txt').touch()
            pathlib.Path(tdir / 'e.txt').touch()
            pathlib.Path(tdir / 'f.json').write_text("{'a': 1}\n")

            path = tdir / 'd.txt'
            for inp, exp in ((str(path), [path]),
                             (path, [path]),
                             (ioinfo_make(path), [ioinfo_make(path)]),
                             (tdir / '*.txt',
                              [tdir / 'd.txt', tdir / 'e.txt']),
                             (tdir.parent / '**' / '*.txt',
                              [tdir / 'd.txt', tdir / 'e.txt']),
                             (tdir.parent / '**' / '*.*',
                              [tdir / 'd.txt',
                               tdir / 'e.txt',
                               tdir / 'f.json']),
                             ([tdir / 'e.txt', tdir / 'd.txt'],
                              [tdir / 'e.txt', tdir / 'd.txt'])
                             ):
                exp = [ioinfo_make(e) for e in exp]
                self.assertEqual(
                    TT.expand_paths(inp), exp, f'{inp!r} vs. {exp!r}'
                )

            with path.open() as fobj:
                res = TT.expand_paths(fobj)
                self.assertEqual(res, [ioinfo_make(fobj)])

# vim:sw=4:ts=4:et:



#
# Forked from m9dicts.tests.{api,dicts}
#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
#
# pylint: disable=missing-docstring,invalid-name
import anyconfig.dicts as TT

from . import common


class TestCase(common.TestCase):
    kind = 'mk_nested_dic'
    pattern = '*.*'
    ordered = False

    def test_mk_nested_dic(self):
        for data in self.each_data():
            val = data.query  # diversion.
            self.assertEqual(
                TT.mk_nested_dic(data.inp, val, **data.opts),
                data.exp
            )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
"""Test cases for anyconfig.parser.parse_single.
"""
import anyconfig.parser as TT

from . import common


class TestCase(common.TestCase):
    kind = 'single'
    pattern = '*.*'

    def test_parse_single(self):
        for data in self.each_data():
            self.assertEqual(TT.parse_single(data.inp), data.exp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=consider-using-with
r"""Module to provide backward compatibility for plugins.
"""
import typing


class BinaryFilesMixin:
    """Mixin class to open configuration files as a binary data.
    """
    _open_flags: typing.Tuple[str, str] = ('rb', 'wb')

    @classmethod
    def ropen(cls, filepath, **kwargs):
        """
        :param filepath: Path to file to open to read data
        """
        return open(filepath, cls._open_flags[0], **kwargs)

    @classmethod
    def wopen(cls, filepath, **kwargs):
        """
        :param filepath: Path to file to open to write data to
        """
        return open(filepath, cls._open_flags[1], **kwargs)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import pathlib
import tempfile
import unittest
import warnings

import anyconfig.schema

from anyconfig.api import ValidationError

from . import common


SCM_NG_0 = '{"type": "object", "properties": {"a": {"type": "string"}}}'


@unittest.skipIf(not anyconfig.schema.SUPPORTED,
                 'jsonschema lib is not available')
class TestCase(common.TestCase):
    kind = 'schema'

    def test_single_load_with_validateion_failures(self):
        with tempfile.TemporaryDirectory() as tdir:
            wdir = pathlib.Path(tdir)
            scm = wdir / 'scm.json'
            scm.write_text(SCM_NG_0)

            for data in self.each_data():
                with warnings.catch_warnings(record=True) as warns:
                    warnings.simplefilter('always')
                    self.assertEqual(
                        self.target_fn(
                            data.inp_path, ac_schema=scm, ac_schema_safe=True,
                            **data.opts
                        ),
                        None
                    )
                    self.assertTrue(len(warns) > 0)
                    self.assertTrue(
                        issubclass(warns[-1].category, UserWarning)
                    )
                    self.assertTrue('scm=' in str(warns[-1].message))

                with self.assertRaises(ValidationError):
                    self.target_fn(
                        data.inp_path, ac_schema=scm, ac_schema_safe=False
                    )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unidiomatic-typecheck
r"""Utility functions for anyconfig.processors.
"""
import operator
import typing
import warnings

import pkg_resources

from ..common import (
    IOInfo, PathOrIOInfoT, UnknownFileTypeError, UnknownProcessorTypeError,
)
from ..ioinfo import make as ioinfo_make
from ..models import processor
from ..utils import concat, groupby
from .datatypes import (
    ProcT, ProcsT, ProcClsT, MaybeProcT
)


def sort_by_prio(prs: typing.Iterable[ProcT]) -> ProcsT:
    """
    Sort an iterable of processor classes by each priority.

    :param prs: A list of :class:`anyconfig.models.processor.Processor` classes
    :return: Sambe as above but sorted by priority
    """
    return sorted(prs, key=operator.methodcaller('priority'), reverse=True)


def select_by_key(items: typing.Iterable[
                    typing.Tuple[typing.List[str], typing.Any]
                  ],
                  sort_fn: typing.Callable[..., typing.Any] = sorted
                  ) -> typing.List[typing.Tuple[str, typing.List[typing.Any]]]:
    """
    :param items: A list of tuples of keys and values, [([key], val)]
    :return: A list of tuples of key and values, [(key, [val])]

    >>> select_by_key([(['a', 'aaa'], 1), (['b', 'bb'], 2), (['a'], 3)])
    [('a', [1, 3]), ('aaa', [1]), ('b', [2]), ('bb', [2])]
    """
    itr = concat(((k, v) for k in ks) for ks, v in items)
    return list((k, sort_fn(t[1] for t in g))
                for k, g in groupby(itr, operator.itemgetter(0)))


def list_by_x(prs: typing.Iterable[ProcT], key: str
              ) -> typing.List[typing.Tuple[str, ProcsT]]:
    """
    :param key: Grouping key, 'type' or 'extensions'
    :return:
        A list of :class:`Processor` or its children classes grouped by
        given 'item', [(cid, [:class:`Processor`)]] by default
    """
    if key == 'type':
        kfn = operator.methodcaller(key)
        res = sorted(((k, sort_by_prio(g)) for k, g
                      in groupby(prs, kfn)),
                     key=operator.itemgetter(0))

    elif key == 'extensions':
        res: typing.List[  # type: ignore
            typing.Tuple[str, ProcsT]
        ] = select_by_key(((p.extensions(), p) for p in prs),
                          sort_fn=sort_by_prio)
    else:
        raise ValueError(
            f"Argument 'key' must be 'type' or 'extensions' but it was '{key}'"
        )

    return res


def findall_with_pred(predicate: typing.Callable[..., bool],
                      prs: ProcsT) -> ProcsT:
    """
    :param predicate: any callable to filter results
    :param prs: A list of :class:`anyconfig.models.processor.Processor` classes
    :return: A list of appropriate processor classes or []
    """
    return sorted((p for p in prs if predicate(p)),
                  key=operator.methodcaller('priority'), reverse=True)


def maybe_processor(type_or_id: typing.Union[ProcT, ProcClsT],
                    cls: ProcClsT = processor.Processor
                    ) -> typing.Optional[ProcT]:
    """
    :param type_or_id:
        Type of the data to process or ID of the processor class or
        :class:`anyconfig.models.processor.Processor` class object or its
        instance
    :param cls: A class object to compare with 'type_or_id'
    :return: Processor instance or None
    """
    if isinstance(type_or_id, cls):
        return type_or_id

    try:
        if issubclass(typing.cast(ProcClsT, type_or_id), cls):
            return type_or_id()  # type: ignore
    except TypeError:
        pass

    return None


def find_by_type_or_id(type_or_id: str, prs: ProcsT) -> ProcsT:
    """
    :param type_or_id: Type of the data to process or ID of the processor class
    :param prs: A list of :class:`anyconfig.models.processor.Processor` classes
    :return:
        A list of processor classes to process files of given data type or
        processor 'type_or_id' found by its ID
    :raises: anyconfig.common.UnknownProcessorTypeError
    """
    def pred(pcls):
        """Predicate"""
        return pcls.cid() == type_or_id or pcls.type() == type_or_id

    pclss = findall_with_pred(pred, prs)
    if not pclss:
        raise UnknownProcessorTypeError(type_or_id)

    return pclss


def find_by_fileext(fileext: str, prs: ProcsT) -> ProcsT:
    """
    :param fileext: File extension
    :param prs: A list of :class:`anyconfig.models.processor.Processor` classes
    :return: A list of processor class to processor files with given extension
    :raises: UnknownFileTypeError
    """
    def pred(pcls):
        """Predicate"""
        return fileext in pcls.extensions()

    pclss = findall_with_pred(pred, prs)
    if not pclss:
        raise UnknownFileTypeError(f'file extension={fileext}')

    return pclss  # :: [Processor], never []


def find_by_maybe_file(obj: PathOrIOInfoT, prs: ProcsT) -> ProcsT:
    """
    :param obj:
        a file path, file or file-like object, pathlib.Path object or an
        'anyconfig.common.IOInfo' (namedtuple) object
    :param cps_by_ext: A list of processor classes
    :return: A list of processor classes to process given (maybe) file
    :raises: UnknownFileTypeError
    """
    if not isinstance(obj, IOInfo):
        obj = ioinfo_make(obj)

    return find_by_fileext(obj.extension, prs)  # :: [Processor], never []


def findall(obj: typing.Optional[PathOrIOInfoT], prs: ProcsT,
            forced_type: typing.Optional[str] = None,
            ) -> ProcsT:
    """
    :param obj:
        a file path, file, file-like object, pathlib.Path object or an
        'anyconfig.common.IOInfo` (namedtuple) object
    :param prs: A list of :class:`anyconfig.models.processor.Processor` classes
    :param forced_type:
        Forced processor type of the data to process or ID of the processor
        class or None

    :return: A list of instances of processor classes to process 'obj' data
    :raises:
        ValueError, UnknownProcessorTypeError,
        UnknownFileTypeError
    """
    if (obj is None or not obj) and forced_type is None:
        raise ValueError(
            "The first argument 'obj' or the second argument 'forced_type' "
            "must be something other than None or False."
        )

    if forced_type is None:
        pclss = find_by_maybe_file(typing.cast(PathOrIOInfoT, obj),
                                   prs)  # :: [Processor], never []
    else:
        pclss = find_by_type_or_id(forced_type, prs)  # Do.

    return pclss


def find(obj: typing.Optional[PathOrIOInfoT], prs: ProcsT,
         forced_type: MaybeProcT = None,
         ) -> ProcT:
    """
    :param obj:
        a file path, file, file-like object, pathlib.Path object or an
        'anyconfig.common.IOInfo' (namedtuple) object
    :param prs: A list of :class:`anyconfig.models.processor.Processor` classes
    :param forced_type:
        Forced processor type of the data to process or ID of the processor
        class or :class:`anyconfig.models.processor.Processor` class object or
        its instance itself
    :param cls: A class object to compare with 'forced_type' later

    :return: an instance of processor class to process 'obj' data
    :raises:
        ValueError, UnknownProcessorTypeError,
        UnknownFileTypeError
    """
    if forced_type is not None and not isinstance(forced_type, str):
        proc = maybe_processor(
            typing.cast(typing.Union[ProcT, ProcClsT], forced_type)
        )
        if proc is None:
            raise ValueError('Wrong processor class or instance '
                             f'was given: {forced_type!r}')

        return proc

    procs = findall(obj, prs, forced_type=typing.cast(str, forced_type))
    return procs[0]


def load_plugins(pgroup: str) -> typing.Iterator[ProcClsT]:
    """
    A generator function to yield a class object of
    :class:`anyconfig.models.processor.Processor`.

    :param pgroup: A string represents plugin type, e.g. anyconfig_backends
    """
    for res in pkg_resources.iter_entry_points(pgroup):
        try:
            yield res.load()
        except ImportError as exc:
            warnings.warn(f'Failed to load plugin, exc={exc!s}')

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
# Suppress import positions after some global variables are defined
# pylint: disable=wrong-import-position
"""A collection of default backend modules.
"""
import warnings

from . import (
    ini,
    json,
    pickle,
    properties,
    shellvars,
    yaml,
    xml
)
from .base import (
    ParserT, ParsersT, ParserClssT
)


PARSERS: ParserClssT = [
    ini.Parser, pickle.Parser, properties.Parser, shellvars.Parser, xml.Parser
] + json.PARSERS

_NA_MSG = "'{}' module is not available. Disabled {} support."

if yaml.PARSERS:
    PARSERS.extend(yaml.PARSERS)
else:
    warnings.warn(_NA_MSG.format('yaml', 'YAML'), ImportWarning)

try:
    from . import toml
    PARSERS.append(toml.Parser)
except ImportError:
    warnings.warn(_NA_MSG.format('toml', 'TOML'), ImportWarning)


__all__ = [
    'ParserT', 'ParsersT', 'ParserClssT',
    'PARSERS',
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring, invalid-name

import pathlib
import unittest

import anyconfig.backend.json
import anyconfig.backend.json.default as JSON
try:
    import anyconfig.backend.yaml.pyyaml as PYYAML
except ImportError:
    PYYAML = None

import anyconfig.parsers.parsers as TT
import anyconfig.ioinfo
import tests.common as TC

from anyconfig.common import (
    UnknownProcessorTypeError, UnknownFileTypeError
)


CNF_PATH = TC.respath('00-cnf.json')


class Test(unittest.TestCase):

    def setUp(self):
        self.psrs = TT.Parsers()

    def test_10_json_parsers(self):
        jpsrs = self.psrs.findall(None, forced_type="json")
        self.assertTrue(isinstance(jpsrs[0], JSON.Parser))

    def test_12_yaml_parsers(self):
        if PYYAML:
            ypsrs = self.psrs.findall(None, forced_type="yaml")
            self.assertTrue(isinstance(ypsrs[0], PYYAML.Parser))

    def test_30_find__ng_cases(self):
        self.assertRaises(ValueError, self.psrs.find, None)
        self.assertRaises(UnknownProcessorTypeError, self.psrs.find, None,
                          forced_type="_unkonw_type_")
        self.assertRaises(UnknownFileTypeError, self.psrs.find,
                          "cnf.unknown_ext")

    def test_32_find__ng_cases(self):
        pcls = anyconfig.backend.json.Parser
        self.assertTrue(isinstance(self.psrs.find("x.conf",
                                                  forced_type="json"),
                                   pcls))
        self.assertTrue(isinstance(self.psrs.find("x.json"), pcls))

        with open(CNF_PATH) as inp:
            self.assertTrue(isinstance(self.psrs.find(inp), pcls))

        if pathlib is not None:
            inp = pathlib.Path("x.json")
            self.assertTrue(isinstance(self.psrs.find(inp), pcls))

    def test_34_find__input_object(self):
        inp = anyconfig.ioinfo.make(CNF_PATH)
        psr = self.psrs.find(inp)
        self.assertTrue(isinstance(psr, anyconfig.backend.json.Parser))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Common parts for YAML backends:
"""
from ...utils import filter_options
from .. import base


def filter_from_options(key, options):
    """
    :param key: Key str in options
    :param options: Mapping object
    :return:
        New mapping object from 'options' in which the item with 'key' filtered

    >>> filter_from_options('a', dict(a=1, b=2))
    {'b': 2}
    """
    return filter_options([k for k in options.keys() if k != key], options)


class Parser(base.StreamParser):
    """
    Parser for YAML files.
    """
    _type = 'yaml'
    _extensions = ['yaml', 'yml']
    _ordered = True
    _allow_primitives = True
    _dict_opts = ['ac_dict']

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring
r"""test cases for anyconfig.utils.lists.
"""
import unittest

import anyconfig.utils.utils as TT


class TestCase(unittest.TestCase):

    def test_filter_options(self):
        data = (
            (('aaa', ), dict(aaa=1, bbb=2), dict(aaa=1)),
            (('aaa', ), dict(bbb=2), dict()),
        )
        for keys, inp, exp in data:
            self.assertEqual(TT.filter_options(keys, inp), exp)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh @ gmmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=invalid-name
r"""ioinfo.constants to provide global constant variables.
"""
import os.path
import re
import typing


GLOB_MARKER: str = '*'
PATH_SEP: str = os.path.sep

SPLIT_PATH_RE: typing.Pattern = re.compile(
    fr'([^{GLOB_MARKER}]+)'
    fr'{PATH_SEP}'
    fr'(.*\{GLOB_MARKER}.*)'
)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2011 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# Ref. python -c "import json; help(json)"
#
# pylint: disable=import-error
r"""JSON backend:

- Format to support: JSON, http://www.json.org
- Requirements: json in python standard library
- Development Status :: 5 - Production/Stable
- Limitations: None obvious
- Special options:

  - All options of json.load{s,} and json.dump{s,} except object_hook
    should work.

  - See also: https://docs.python.org/3/library/json.html or
    https://docs.python.org/2/library/json.html

Changelog:

.. versionchanged:: 0.9.8

   - Moved from ..json.py
   - Drop simplejson support from this module

.. versionchanged:: 0.9.6

   - Add support of loading primitives other than mapping objects.

.. versionadded:: 0.0.1
"""
import json

from .. import base
from .common import Parser as BaseParser


class Parser(BaseParser):
    """
    Parser for JSON files.
    """
    _cid = 'std.json'
    _priority = 30  # Higher priority than others.

    _load_from_string_fn = base.to_method(json.loads)
    _load_from_stream_fn = base.to_method(json.load)
    _dump_to_string_fn = base.to_method(json.dumps)
    _dump_to_stream_fn = base.to_method(json.dump)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

from .. import base


class Collector(base.TDataCollector):
    ordered: bool = True


class TestCase(unittest.TestCase, Collector):

    def setUp(self):
        self.init()

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Common functions and variables.
"""
import typing

from ..models import processor


ProcT = typing.TypeVar('ProcT', bound=processor.Processor)
ProcsT = typing.List[ProcT]
ProcClsT = typing.Type[ProcT]
ProcClssT = typing.List[ProcClsT]

MaybeProcT = typing.Optional[typing.Union[str, ProcT, ProcClsT]]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2018 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""A class collect anyconfig.models.processor.Processor and inherited class
objects inherited from it.
"""
import operator
import typing

from ..common import PathOrIOInfoT
from . import utils
from .datatypes import (
    ProcT, ProcsT, ProcClsT, ProcClssT, MaybeProcT
)


class Processors:
    """An abstract class of which instance holding processors.
    """
    _pgroup: str = ''  # processor group name to load plugins

    def __init__(self, processors: typing.Optional[ProcClssT] = None) -> None:
        """
        :param processors:
            A list of :class:`anyconfig.models.processor.Processor` or its
            children class objects to initialize this, or None
        """
        # {<processor_class_id>: <processor_instance>}
        self._processors: typing.Dict[str, ProcT] = dict()  # type: ignore
        if processors is not None:
            for pcls in processors:
                self.register(pcls)

        self.load_plugins()

    def register(self, pcls: ProcClsT) -> None:
        """
        :param pclss: :class:`Processor` or its children class objects
        """
        if pcls.cid() not in self._processors:
            self._processors[pcls.cid()] = pcls()

    def load_plugins(self) -> None:
        """Load and register pluggable processor classes internally.
        """
        if self._pgroup:
            for pcls in utils.load_plugins(self._pgroup):
                self.register(pcls)

    def list(self, sort: bool = False) -> ProcClssT:
        """
        :param sort: Result will be sorted if it's True
        :return: A list of :class:`Processor` or its children classes
        """
        prs = self._processors.values()
        if sort:
            return sorted(prs, key=operator.methodcaller('cid'))

        return list(prs)

    def list_by_cid(self) -> typing.List[typing.Tuple[str, ProcsT]]:
        """
        :return:
            A list of :class:`Processor` or its children classes grouped by
            each cid, [(cid, [:class:`Processor`)]]
        """
        prs = self._processors
        return sorted(((cid, [prs[cid]]) for cid in sorted(prs.keys())),
                      key=operator.itemgetter(0))

    def list_by_type(self) -> typing.List[typing.Tuple[str, ProcsT]]:
        """
        :return:
            A list of :class:`Processor` or its children classes grouped by
            each type, [(type, [:class:`Processor`)]]
        """
        return utils.list_by_x(self.list(), 'type')  # type: ignore

    def list_by_x(self, item: typing.Optional[str] = None
                  ) -> typing.List[typing.Tuple[str, ProcsT]]:
        """
        :param item: Grouping key, one of 'cid', 'type' and 'extensions'
        :return:
            A list of :class:`Processor` or its children classes grouped by
            given 'item', [(cid, [:class:`Processor`)]] by default
        """
        prs = self._processors

        if item is None or item == 'cid':  # Default.
            res = [(cid, [prs[cid]]) for cid in sorted(prs.keys())]

        elif item in ('type', 'extensions'):
            res = utils.list_by_x(prs.values(), typing.cast(str, item))
        else:
            raise ValueError("keyword argument 'item' must be one of "
                             "None, 'cid', 'type' and 'extensions' "
                             f"but it was '{item}'")
        return res

    def list_x(self, key: typing.Optional[str] = None) -> typing.List[str]:
        """
        :param key: Which of key to return from 'cid', 'type', and 'extention'
        :return: A list of x 'key'
        """
        if key in ('cid', 'type'):
            return sorted(set(operator.methodcaller(key)(p)
                              for p in self._processors.values()))
        if key == 'extension':
            return sorted(k for k, _v in self.list_by_x('extensions'))

        raise ValueError("keyword argument 'key' must be one of "
                         "None, 'cid', 'type' and 'extension' "
                         f"but it was '{key}'")

    def findall(self, obj: typing.Optional[PathOrIOInfoT],
                forced_type: typing.Optional[str] = None
                ) -> typing.List[ProcT]:
        """
        :param obj:
            a file path, file, file-like object, pathlib.Path object or an
            'anyconfig.common.IOInfo' (namedtuple) object
        :param forced_type: Forced processor type to find

        :return: A list of instances of processor classes to process 'obj'
        :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
        """
        return utils.findall(obj, self.list(), forced_type=forced_type)

    def find(self, obj: typing.Optional[PathOrIOInfoT],
             forced_type: MaybeProcT = None) -> ProcT:
        """
        :param obj:
            a file path, file, file-like object, pathlib.Path object or an
            'anyconfig.common.IOInfo' (namedtuple) object
        :param forced_type:
            Forced processor type to find or a processor class object or a
            processor intance

        :return: an instance of processor class to process 'obj'
        :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
        """
        return utils.find(obj, self.list(), forced_type=forced_type)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2015 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Implementation using jsonschema provides the following functions.

- validate(data: typing.Dict[str, typing.Any],
           schema: typing.Dict[str, typing.Any],
           ac_schema_safe: bool = True, ac_schema_errors: bool = False,
           **options) -> typing.Tuple[bool, str]:
  validate with schema

- gen_schema(data: typing.Dict[str, typing.Any],
             **options) -> typing.Dict[str, typing.Any]:
  Generate an object represents a schema
"""
import typing
import warnings

import jsonschema

from ..common import (
    ValidationError, InDataExT, InDataT
)
from ..utils import (
    filter_options, is_dict_like, is_list_like
)
from .datatypes import ResultT


def _validate_all(data: InDataExT, schema: InDataT, **_options) -> ResultT:
    """
    See the description of :func:`validate` for more details of parameters and
    return value.

    :seealso: https://python-jsonschema.readthedocs.io/en/latest/validate/,
    a section of 'iter_errors' especially
    """
    vldtr = jsonschema.Draft7Validator(schema)  # :raises: SchemaError, ...
    errors = list(vldtr.iter_errors(data))

    return (not errors, [err.message for err in errors])


def _validate(data: InDataExT, schema: InDataT, ac_schema_safe: bool = True,
              **options: typing.Any) -> ResultT:
    """
    See the description of :func:`validate` for more details of parameters and
    return value.

    Validate target object 'data' with given schema object.
    """
    try:
        jsonschema.validate(
            data, schema, format_checker=jsonschema.draft7_format_checker,
            **options
        )
    except (jsonschema.ValidationError, jsonschema.SchemaError,
            Exception) as exc:
        if ac_schema_safe:
            return (False, str(exc))  # Validation was failed.
        raise

    return (True, '')


def validate(data: InDataExT, schema: InDataT, ac_schema_safe: bool = True,
             ac_schema_errors: bool = False, **options: typing.Any
             ) -> ResultT:
    """
    Validate target object with given schema object, loaded from JSON schema.

    See also: https://python-jsonschema.readthedocs.org/en/latest/validate/

    :parae data: Target object (a dict or a dict-like object) to validate
    :param schema: Schema object (a dict or a dict-like object)
        instantiated from schema JSON file or schema JSON string
    :param options: Other keyword options such as:

        - ac_schema_safe: Exception (jsonschema.ValidationError or
          jsonschema.SchemaError or others) will be thrown during validation
          process due to any validation or related errors. However, these will
          be catched by default, and will be re-raised if this value is False.

        - ac_schema_errors: Lazily yield each of the validation errors and
          returns all of them if validation fails.

    :return: (True if validation succeeded else False, error message[s])
    """
    options = filter_options(('cls', ), options)
    if ac_schema_errors:
        return _validate_all(data, schema, **options)

    return _validate(data, schema, ac_schema_safe, **options)


def is_valid(data: InDataExT, schema: InDataT, ac_schema_safe: bool = True,
             ac_schema_errors: bool = False, **options) -> bool:
    """
    Raise ValidationError if data `data` was invalidated by schema `schema`.
    """
    if schema is None or not schema:
        return True

    (_success, error_or_errors) = validate(
        data, schema, ac_schema_safe=True,
        ac_schema_errors=ac_schema_errors, **options
    )
    if error_or_errors:
        msg = f'scm={schema!s}, err={error_or_errors!s}'
        if ac_schema_safe:
            warnings.warn(msg)
            return False

        raise ValidationError(msg)

    return True


_SIMPLETYPE_MAP: typing.Dict[typing.Any, str] = {
    list: 'array', tuple: 'array', bool: 'boolean', int: 'integer', float:
    'number', dict: 'object', str: 'string'
}


def _process_options(**options):
    """
    Helper function to process keyword arguments passed to gen_schema.

    :return: A tuple of (typemap :: dict, strict :: bool)
    """
    return (options.get('ac_schema_typemap', _SIMPLETYPE_MAP),
            bool(options.get('ac_schema_strict', False)))


def array_to_schema(iarr: typing.Iterable[InDataT], **options
                    ) -> typing.Dict[str, typing.Any]:
    """
    Generate a JSON schema object with type annotation added for given object.

    :param arr: Array of mapping objects like dicts
    :param options: Other keyword options such as:

        - ac_schema_strict: True if more strict (precise) schema is needed
        - ac_schema_typemap: Type to JSON schema type mappings

    :return: Another mapping objects represents JSON schema of items
    """
    (typemap, strict) = _process_options(**options)

    arr: typing.List[InDataT] = list(iarr)
    scm = {
        'type': typemap[list],
        'items': gen_schema(arr[0] if arr else 'str', **options)
    }
    if strict:
        nitems = len(arr)
        scm['minItems'] = nitems
        scm['uniqueItems'] = len(set(arr)) == nitems

    return scm


def object_to_schema(obj: InDataT, **options) -> InDataT:
    """
    Generate a node represents JSON schema object with type annotation added
    for given object node.

    :param obj: mapping object such like a dict
    :param options: Other keyword options such as:

        - ac_schema_strict: True if more strict (precise) schema is needed
        - ac_schema_typemap: Type to JSON schema type mappings

    :yield: Another mapping objects represents JSON schema of object
    """
    (typemap, strict) = _process_options(**options)

    props = dict((k, gen_schema(v, **options)) for k, v in obj.items())
    scm = {'type': typemap[dict], 'properties': props}
    if strict:
        scm['required'] = sorted(props.keys())

    return scm


_SIMPLE_TYPES = (bool, int, float, str)


def gen_schema(data: InDataExT, **options) -> InDataT:
    """
    Generate a node represents JSON schema object with type annotation added
    for given object node.

    :param data: Configuration data object (dict[-like] or namedtuple)
    :param options: Other keyword options such as:

        - ac_schema_strict: True if more strict (precise) schema is needed
        - ac_schema_typemap: Type to JSON schema type mappings

    :return: A dict represents JSON schema of this node
    """
    if data is None:
        return {'type': 'null'}

    _type = type(data)

    if _type in _SIMPLE_TYPES:
        typemap = options.get('ac_schema_typemap', _SIMPLETYPE_MAP)
        scm = {'type': typemap[_type]}

    elif is_dict_like(data):
        scm = object_to_schema(data, **options)  # type: ignore

    elif is_list_like(data):
        scm = array_to_schema(
            typing.cast(typing.Iterable[InDataT], data), **options
        )

    return scm

# vim:sw=4:ts=4:et:



"""setup.py to build package.
"""
import os
import pathlib
import re
import setuptools
import setuptools.command.bdist_rpm


# It might throw IndexError and so on.
VERSION = '0.1.0'
VER_REG = re.compile(r"^__version__ = '([^']+)'")

for fpath in pathlib.Path('src').glob('**/__init__.py'):
    for line in fpath.open():
        match = VER_REG.match(line)
        if match:
            VERSION = match.groups()[0]
            break

# For daily snapshot versioning mode:
RELEASE = "1%{?dist}"
if os.environ.get("_SNAPSHOT_BUILD", None) is not None:
    import datetime
    RELEASE = RELEASE.replace('1',
                              datetime.datetime.now().strftime("%Y%m%d"))


def _replace(line):
    """Replace some strings in the RPM SPEC template"""
    if "@VERSION@" in line:
        return line.replace("@VERSION@", VERSION)

    if "@RELEASE@" in line:
        return line.replace("@RELEASE@", RELEASE)

    if "Source0:" in line:  # Dirty hack
        return "Source0: %{pkgname}-%{version}.tar.gz"

    return line


class bdist_rpm(setuptools.command.bdist_rpm.bdist_rpm):
    """Override the default content of the RPM SPEC.
    """
    spec_tmpl = pathlib.Path('pkg/package.spec.in').resolve()

    def _make_spec_file(self):
        return [_replace(line.rstrip()) for line in self.spec_tmpl.open()]


setuptools.setup(version=VERSION, cmdclass=dict(bdist_rpm=bdist_rpm),
                 data_files=[("share/man/man1", ["docs/anyconfig_cli.1"])])

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2015 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-argument
"""Default (dummy) implementation.
"""
import typing

from ..common import (
    InDataT, InDataExT
)
from .datatypes import ResultT


def validate(data: InDataExT, schema: InDataT, ac_schema_safe: bool = True,
             ac_schema_errors: bool = False, **options: typing.Any
             ) -> ResultT:
    """
    Dummy function does not validate at all in actual.
    """
    return (True, 'Validation module (jsonschema) is not available')


def is_valid(data: InDataExT, schema: InDataT, ac_schema_safe: bool = True,
             ac_schema_errors: bool = False, **options) -> bool:
    """
    Dummy function never raise exceptions.
    """
    return True


def gen_schema(data: InDataExT, **options) -> InDataT:
    """
    Dummy function generates an empty dict in actual.
    """
    return dict()


# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=unused-import,import-error,invalid-name
r"""Internal APIs to load, list and find parser class objects.
"""
import typing

from ..common import PathOrIOInfoT
from ..backend import ParserT, ParsersT
from .parsers import Parsers


MaybeParserT = typing.Optional[
    typing.Union[str, ParserT, typing.Type[ParserT]]
]


def load_plugins() -> None:
    """[Re-]Load pluggable processors.
    """
    Parsers().load_plugins()


def list_types() -> typing.List[str]:
    """List supported processor types.
    """
    return sorted(Parsers().list_x('type'))


def list_by_cid() -> typing.List[typing.Tuple[str, ParsersT]]:
    """
    List processors by each cid.
    """
    return Parsers().list_by_x('cid')


def list_by_type() -> typing.List[typing.Tuple[str, ParsersT]]:
    """
    List processor by eacch type.
    """
    return Parsers().list_by_x('type')


def list_by_extension() -> typing.List[typing.Tuple[str, ParsersT]]:
    """
    List processor by file extension supported.
    """
    return Parsers().list_by_x('extensions')


def findall(obj: typing.Optional[PathOrIOInfoT] = None,
            forced_type: typing.Optional[str] = None
            ) -> typing.List[ParserT]:
    """
    Find out processor objects can process data from given 'obj' which may be a
    file path, file or file-like object, pathlib.Path object or an
    'anyconfig.common.IOInfo' (namedtuple) object.

    :param obj:
        a file path, file or file-like object, pathlib.Path object, an
        'anyconfig.common.IOInfo' (namedtuple) object, or None
    :param forced_type: Forced type or id of the processor

    :return: A list of instances of processor classes to process 'obj'
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    return Parsers().findall(obj, forced_type=forced_type)


def find(obj: typing.Optional[PathOrIOInfoT] = None,
         forced_type: MaybeParserT = None) -> ParserT:
    """
    This function is very similar to the above :func:`findall` but returns
    *a processor object* instead of a list of processor objects.

    :param obj:
        a file path, file or file-like object, pathlib.Path object, an
        'anyconfig.common.IOInfo' (namedtuple) object, or None
    :param forced_type: Forced type or id of the processor

    :return:
        An instance of processor class of highest priority to process 'obj'
    :raises: ValueError, UnknownProcessorTypeError, UnknownFileTypeError
    """
    return Parsers().find(obj, forced_type=forced_type)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import anyconfig.api._dump as TT

from anyconfig.api import UnknownProcessorTypeError

from . import common


class TestCase(common.BaseTestCase):

    def test_dumps(self):
        for data in self.each_data():
            self.assertEqual(
                TT.dumps(data.inp, **data.opts).strip(),
                data.exp.strip(),
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_dumps_intentional_failures(self):
        for data in self.each_data():
            with self.assertRaises(AssertionError):
                self.assertEqual(TT.dumps(data.inp, **data.opts).strip(), {})

    def test_dump_failure_ac_parser_was_not_given(self):
        for data in self.each_data():
            with self.assertRaises(ValueError):
                TT.dumps(data.inp)

    def test_dump_failure_invalid_ac_parser_was_given(self):
        for data in self.each_data():
            with self.assertRaises(UnknownProcessorTypeError):
                TT.dumps(data.inp, ac_parser='invalid_id')

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# pylint: disable=missing-docstring,invalid-name
r"""test cases for anyconfig.utils.
"""
import collections
import pathlib
import unittest
import typing

import anyconfig.ioinfo
import anyconfig.utils.detectors as TT


class TestCase(unittest.TestCase):

    def _run(self,
             target_fn: typing.Callable[..., typing.Any],
             ies: typing.Iterable[typing.Tuple[typing.Any, bool]]
             ) -> None:
        for inp, exp in ies:
            meth = self.assertTrue if exp else self.assertFalse
            meth(target_fn(inp), f'input: {inp!r}, expected: {exp!r}')

    def test_is_iterable(self):
        self._run(
            TT.is_iterable,
            (
             (None, False),
             ([], True), ((), True),
             ((str(x) for x in range(10)), True),
             ([str(x) for x in range(10)], True),
             ('abc', False), (0, False), ({}, False),
             )
        )

    def test_is_list_like(self):
        self._run(
            TT.is_list_like,
            (
             (None, False),
             (0, False),
             ('aaa', False),
             ({}, False),
             ([], True), ((), True),
             ((str(x) for x in range(10)), True),
             ([str(x) for x in range(10)], True),
             )
        )

    def test_is_dict_like(self):
        self._run(
            TT.is_dict_like,
            (
             (None, False),
             (0, False),
             ('aaa', False),
             ([], False),
             ((1, ), False),
             (collections.namedtuple('Point', ('x', 'y'))(1, 2), False),
             ({}, True),
             (collections.OrderedDict((('a', 1), ('b', 2))), True),
             )
        )

    def test_is_path(self):
        self._run(
            TT.is_path,
            (
             (None, False),
             ('/tmp/t.xt', True),
             (0, False),
             (pathlib.Path(__file__), False),
             (open(__file__), False),
             (anyconfig.ioinfo.make(__file__), False),
             )
        )

    def test_is_path_obj(self):
        self._run(
            TT.is_path_obj,
            (
             (None, False),
             (__file__, False),
             (pathlib.Path(__file__), True),
             (str(pathlib.Path(__file__).resolve()), False),
             (open(__file__), False),
             (anyconfig.ioinfo.make(__file__), False),
             )
        )

    def test_is_file_stream(self):
        self._run(
            TT.is_file_stream,
            (
             (None, False),
             (__file__, False),
             (pathlib.Path(__file__), False),
             (str(pathlib.Path(__file__).resolve()), False),
             (open(__file__), True),
             (anyconfig.ioinfo.make(__file__), False),
             )
        )

    def test_is_ioinfo(self):
        self._run(
            TT.is_ioinfo,
            (
             (None, False),
             (__file__, False),
             (pathlib.Path(__file__), False),
             (str(pathlib.Path(__file__).resolve()), False),
             (open(__file__), False),
             (anyconfig.ioinfo.make(__file__), True),
             (anyconfig.ioinfo.make(open(__file__)), True),
             )
        )

    def test_is_stream_ioinfo(self):
        self._run(
            TT.is_stream_ioinfo,
            (
             (None, False),
             (__file__, False),
             (pathlib.Path(__file__), False),
             (str(pathlib.Path(__file__).resolve()), False),
             (open(__file__), False),
             (anyconfig.ioinfo.make(__file__), False),
             (anyconfig.ioinfo.make(open(__file__)), True),
             )
        )

    def test_is_path_like_object(self):
        self._run(
            TT.is_path_like_object,
            (
             (False, False),
             (None, False),
             (__file__, True),
             (pathlib.Path(__file__), True),
             (str(pathlib.Path(__file__).resolve()), True),
             (open(__file__), True),
             (anyconfig.ioinfo.make(__file__), True),
             (anyconfig.ioinfo.make(open(__file__)), True),
             )
        )

    def test_is_paths(self):
        self._run(
            TT.is_paths,
            (
             (False, False),
             (None, False),
             (__file__, False),
             (pathlib.Path(__file__), False),
             (str(pathlib.Path(__file__).resolve()), False),
             (open(__file__), False),
             (anyconfig.ioinfo.make(__file__), False),
             (anyconfig.ioinfo.make(open(__file__)), False),
             ('/tmp/*.txt', True),
             (pathlib.Path('/tmp/*.txt'), True),
             ((__file__, ), True),
             ((pathlib.Path(__file__), ), True),
             ((str(pathlib.Path(__file__).resolve()), ), True),
             ((open(__file__), ), True),
             ((anyconfig.ioinfo.make(__file__), ), True),
             ((anyconfig.ioinfo.make(open(__file__)), ), True),
             )
        )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

import anyconfig.template

from . import common


@unittest.skipIf(not anyconfig.template.SUPPORTED,
                 'jinja2 template lib is not available')
class TestCase(common.TestCase):
    kind = 'template'

    def test_multi_load(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn(
                    tdata.inputs, ac_context=tdata.ctx, **tdata.opts
                ),
                tdata.exp,
                tdata
            )

    def test_multi_load_failures(self):
        for tdata in self.each_data():
            with self.assertRaises(AssertionError):
                self.assertEqual(
                    self.target_fn(
                        tdata.inputs, ac_context=tdata.ctx, **tdata.opts
                    ),
                    None,
                    tdata
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Utility functions to list and find appropriate parser class objects and
instances.

.. versionchanged:: 0.10.2

   - Split and re-organize the module and export only some functions.

.. versionadded:: 0.9.5

   - Add to abstract processors such like Parsers (loaders and dumpers).
"""
from .utils import (
    load_plugins, list_types, list_by_cid, list_by_type, list_by_extension,
    findall, find, MaybeParserT
)

__all__ = [
    'load_plugins', 'list_types', 'list_by_cid', 'list_by_type',
    'list_by_extension', 'findall', 'find', 'MaybeParserT'
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest
import warnings

import anyconfig.api._load as TT
import anyconfig.template

from . import common


@unittest.skipIf(not anyconfig.template.SUPPORTED,
                 'jinja2 template lib is not available')
class TestCase(common.TestCase):
    kind = 'template'

    def test_loads_template(self):
        for data in self.each_data():
            self.assertEqual(
                TT.loads(data.inp, ac_context=data.ctx, **data.opts),
                data.exp,
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_loads_from_template_failures(self):
        inp = '{"a": "{{ a"}'
        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter('always')
            res = TT.loads(inp, ac_parser='json', ac_template=True)
            self.assertEqual(res, dict(a='{{ a'))
            # self.assertEqual(len(warns), 1)  # Needs to fix plugins
            self.assertTrue(issubclass(warns[-1].category, UserWarning))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2019 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring, invalid-name, no-member
import pathlib
import pickle
import tempfile

import anyconfig.api._open as TT
import anyconfig.api._load as LD

from . import common


class TestCase(common.BaseTestCase):

    def test_open_text_io(self):
        for data in self.each_data():
            with TT.open(data.inp_path, **data.opts) as inp:
                self.assertEqual(LD.loads(inp.read(), **data.opts), data.inp)

    def test_open_byte_io(self):
        cnf = dict(a=1, b='b')

        with tempfile.TemporaryDirectory() as workdir:
            path = pathlib.Path(workdir) / 'test.pkl'
            pickle.dump(cnf, path.open(mode='wb'))

            with TT.open(path) as fio:
                self.assertEqual(fio.mode, 'rb')
                self.assertEqual(
                    LD.loads(fio.read(), ac_parser='pickle'),
                    LD.load(path)
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Utility funtions for anyconfig.ionfo.
"""
import pathlib
import typing

from .constants import (
    GLOB_MARKER, PATH_SEP, SPLIT_PATH_RE
)


def is_io_stream(obj: typing.Any) -> bool:
    """Is given object ``obj`` an IO stream, file or file-like object?
    """
    return callable(getattr(obj, 'read', False))


def get_path_and_ext(path: pathlib.Path) -> typing.Tuple[pathlib.Path, str]:
    """Normaliez path objects and retunr it with file extension.
    """
    abs_path = path.expanduser().resolve()
    file_ext = abs_path.suffix
    return (
        abs_path,
        file_ext[1:] if file_ext.startswith('.') else ''
    )


def split_path_by_marker(path: str,
                         marker: str = GLOB_MARKER,
                         path_sep: str = PATH_SEP,
                         split_re: typing.Pattern = SPLIT_PATH_RE
                         ) -> typing.Tuple[str, str]:
    """Split given path string by the marker.
    """
    if marker not in path:
        return (path, '')

    if path_sep not in path:
        return ('', path)

    matched = split_re.match(path)
    if not matched:
        raise ValueError(f'Invalid path: {path}')

    return typing.cast(typing.Tuple[str, str], matched.groups())

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2015 - 2019 Satoru SATOH <satoru.satoh@gmail.com>
# Copyright (C) 2017 Red Hat, Inc.
# License: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods
# pylint: disable=ungrouped-imports
from collections import OrderedDict as ODict

try:
    import anyconfig.backend.toml as TT
except ImportError:
    import unittest
    raise unittest.SkipTest

import tests.backend.common as TBC


_DOB = TT.toml.loads("dob = 1979-05-27T07:32:00Z")['dob']

CNF = ODict((('title', 'TOML Example'),
             ('owner',
              ODict((('name', 'Tom Preston-Werner'),
                     ('dob', _DOB)))),
             ('database',
              ODict((('server', '192.168.1.1'),
                     ('ports', [8001, 8001, 8002]),
                     ('connection_max', 5000),
                     ('enabled', True)))),
             ('servers',
              ODict((('alpha',
                      ODict((('ip', '10.0.0.1'), ('dc', 'eqdc10')))),
                     ('beta',
                      ODict((('ip', '10.0.0.2'), ('dc', 'eqdc10'))))))),
             ('clients',
              ODict((('data', [['gamma', 'delta'], [1, 2]]),
                     ('hosts', ['alpha', 'omega']))))))


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf = CNF
    cnf_s = TBC.read_from_res("20-00-cnf.toml")


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):

    load_options = dump_options = dict(dummy="this_will_be_ignored")


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):

    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
"""Utility funtions for anyconfig.ionfo.
"""
import typing

if typing.TYPE_CHECKING:
    from .. import common


def are_same_file_types(objs: typing.List['common.IOInfo']) -> bool:
    """
    Are given objects, pathlib.Path or io, same type (have same extension)?
    """
    if not objs:
        return False

    ext = objs[0].extension
    return all(p.extension == ext for p in objs[1:])

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Utility functions in anyconfig.backend.base.
"""
import typing

from ...common import (
    InDataT, IOInfo, PathOrIOInfoT, InDataExT
)


OutDataExT = InDataExT

IoiT = IOInfo
MaybeFilePathT = typing.Optional[PathOrIOInfoT]

GenContainerT = typing.Callable[..., InDataT]
OptionsT = typing.Dict[str, typing.Any]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring,invalid-name,too-few-public-methods
# pylint: disable=ungrouped-imports
import anyconfig.backend.json.default as TT
import tests.backend.common as TBC

from collections import OrderedDict


CNF_0_S = TBC.read_from_res("20-00-cnf.json")
CNF_0 = OrderedDict((("a", 0), ("b", "bbb"), ("c", 5),
                     ("sect0", OrderedDict((("d", ["x", "y", "z"]), )))))


class HasParserTrait(TBC.HasParserTrait):

    psr = TT.Parser()
    cnf_s = CNF_0_S
    cnf = CNF_0


class Test_10(TBC.Test_10_dumps_and_loads, HasParserTrait):

    load_options = dump_options = dict(parse_int=None, indent=2)
    empty_patterns = [('', {}), ('{}', {}), ('[]', []), ('null', None)]


class Test_20(TBC.Test_20_dump_and_load, HasParserTrait):
    pass

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Common test Utility functions, constants and global variables, etc.
"""
from .constants import TESTS_DIR, RES_DIR, NULL_CNTNR
from .collector import TDataCollector
from .datatypes import TDataPaths, TData
from .utils import (
    load_data, each_data_from_dir, maybe_data_path
)


__all__ = [
    'TESTS_DIR', 'RES_DIR', 'NULL_CNTNR',
    'TDataCollector',
    'TDataPaths', 'TData',
    'load_data', 'each_data_from_dir', 'maybe_data_path'
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Public API to query data with JMESPath expression.
"""
try:
    from .query import try_query
    SUPPORTED = True
except ImportError:
    from .default import try_query
    SUPPORTED = False  # type: ignore


__all__ = [
    'try_query',
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh @ gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Abstract and basic dumpes.
"""
import io
import typing

from ...utils import (
    filter_options, is_stream_ioinfo
)
from .datatypes import (
    InDataExT, IoiT
)
from .utils import (
    ensure_outdir_exists, not_implemented
)


class DumperMixin:
    """
    Mixin class to dump data.

    Inherited classes must implement the following methods.

    - :meth:`dump_to_string`: Dump config as a string
    - :meth:`dump_to_stream`: Dump config to a file or file-like object
    - :meth:`dump_to_path`: Dump config to a file of given path

    Member variables:

    - _dump_opts: Backend specific options on dump
    - _open_write_mode: Backend option to specify write mode passed to open()
    """
    _dump_opts: typing.List[str] = []
    _open_write_mode = 'w'

    def wopen(self, filepath: str, **kwargs):
        """
        Open file ``filepath`` with the write mode ``_open_write_mode``.
        """
        return open(  # pylint: disable=consider-using-with
            filepath, self._open_write_mode, **kwargs
        )

    def dump_to_string(self, cnf: InDataExT, **kwargs) -> str:
        """
        Dump config 'cnf' to a string.

        :param cnf: Configuration data to dump
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: string represents the configuration
        """
        not_implemented(self, cnf, **kwargs)
        return ''

    def dump_to_path(self, cnf: InDataExT, filepath: str, **kwargs) -> None:
        """
        Dump config 'cnf' to a file 'filepath'.

        :param cnf: Configuration data to dump
        :param filepath: Config file path
        :param kwargs: optional keyword parameters to be sanitized :: dict
        """
        not_implemented(self, cnf, filepath, **kwargs)

    def dump_to_stream(self, cnf: InDataExT, stream: typing.IO, **kwargs
                       ) -> None:
        """
        Dump config 'cnf' to a file-like object 'stream'.

        TODO: How to process socket objects same as file objects ?

        :param cnf: Configuration data to dump
        :param stream:  Config file or file like object
        :param kwargs: optional keyword parameters to be sanitized :: dict
        """
        not_implemented(self, cnf, stream, **kwargs)

    def dumps(self, cnf: InDataExT, **kwargs) -> str:
        """
        Dump config 'cnf' to a string.

        :param cnf: Configuration data to dump
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: string represents the configuration
        """
        kwargs = filter_options(self._dump_opts, kwargs)
        return self.dump_to_string(cnf, **kwargs)

    def dump(self, cnf: InDataExT, ioi: IoiT, **kwargs):
        """
        Dump config 'cnf' to output object of which 'ioi' referring.

        :param cnf: Configuration data to dump
        :param ioi:
            an 'anyconfig.cmmon.IOInfo' namedtuple object provides various
            info of input object to load data from

        :param kwargs: optional keyword parameters to be sanitized :: dict
        :raises IOError, OSError, AttributeError: When dump failed.
        """
        kwargs = filter_options(self._dump_opts, kwargs)

        if is_stream_ioinfo(ioi):
            self.dump_to_stream(cnf, typing.cast(typing.IO, ioi.src), **kwargs)
        else:
            ensure_outdir_exists(ioi.path)
            self.dump_to_path(cnf, ioi.path, **kwargs)


class BinaryDumperMixin(DumperMixin):
    """
    Mixin class to dump binary (byte string) configuration data.
    """
    _open_write_mode: str = 'wb'


class ToStringDumperMixin(DumperMixin):
    """
    Abstract config parser provides a method to dump configuration to a file or
    file-like object (stream) and a file of given path to help implement parser
    of which backend lacks of such functions.

    Parser classes inherit this class have to override the method
    :meth:`dump_to_string` at least.
    """
    def dump_to_path(self, cnf: InDataExT, filepath: str, **kwargs) -> None:
        """
        Dump config 'cnf' to a file 'filepath'.

        :param cnf: Configuration data to dump
        :param filepath: Config file path
        :param kwargs: optional keyword parameters to be sanitized :: dict
        """
        with self.wopen(filepath) as out:
            out.write(self.dump_to_string(cnf, **kwargs))

    def dump_to_stream(self, cnf: InDataExT, stream: typing.IO, **kwargs
                       ) -> None:
        """
        Dump config 'cnf' to a file-like object 'stream'.

        TODO: How to process socket objects same as file objects ?

        :param cnf: Configuration data to dump
        :param stream:  Config file or file like object
        :param kwargs: optional keyword parameters to be sanitized :: dict
        """
        stream.write(self.dump_to_string(cnf, **kwargs))


class ToStreamDumperMixin(DumperMixin):
    """
    Abstract config parser provides methods to dump configuration to a string
    content or a file of given path to help implement parser of which backend
    lacks of such functions.

    Parser classes inherit this class have to override the method
    :meth:`dump_to_stream` at least.
    """
    def dump_to_string(self, cnf: InDataExT, **kwargs) -> str:
        """
        Dump config 'cnf' to a string.

        :param cnf: Configuration data to dump
        :param kwargs: optional keyword parameters to be sanitized :: dict

        :return: Dict-like object holding config parameters
        """
        stream = io.StringIO()
        self.dump_to_stream(cnf, stream, **kwargs)
        return stream.getvalue()

    def dump_to_path(self, cnf: InDataExT, filepath: str, **kwargs) -> None:
        """
        Dump config 'cnf' to a file 'filepath`.

        :param cnf: Configuration data to dump
        :param filepath: Config file path
        :param kwargs: optional keyword parameters to be sanitized :: dict
        """
        with self.wopen(filepath) as out:
            self.dump_to_stream(cnf, out, **kwargs)

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
# For 'anyconfig.open':
# pylint: disable=redefined-builtin
r"""
.. module:: anyconfig
   :platform: Unix, Windows
   :synopsis: Generic interface to loaders for various config file formats.

python-anyconfig is a `MIT licensed <http://opensource.org/licenses/MIT>`_
python library provides common APIs to access to configuration files in various
formats with some useful features such as contents merge, templates and schema
validation/generation support.

- Home: https://github.com/ssato/python-anyconfig
- (Latest) Doc: http://python-anyconfig.readthedocs.org/en/latest/
- PyPI: https://pypi.python.org/pypi/anyconfig
- Copr RPM repos: https://copr.fedoraproject.org/coprs/ssato/python-anyconfig/

"""
from .api import (
    dump, dumps, single_load, multi_load, load, loads,
    open, version,
    UnknownFileTypeError, UnknownParserTypeError,
    UnknownProcessorTypeError, ValidationError,
    MS_REPLACE, MS_NO_REPLACE, MS_DICTS, MS_DICTS_AND_LISTS, MERGE_STRATEGIES,
    merge, get, set_,
    load_plugins, list_types, list_by_cid, list_by_type, list_by_extension,
    findall, find,
    try_query,
    validate, is_valid, gen_schema
)


__all__ = [
    'dump', 'dumps',
    'single_load', 'multi_load', 'load', 'loads',
    'open', 'version',

    # anyconfig.common
    'UnknownParserTypeError', 'UnknownProcessorTypeError',
    'UnknownFileTypeError', 'ValidationError',

    # anyconfig.dicsts
    'MS_REPLACE', 'MS_NO_REPLACE', 'MS_DICTS', 'MS_DICTS_AND_LISTS',
    'MERGE_STRATEGIES', 'merge', 'get', 'set_',

    # anyconfig.parsers
    'load_plugins', 'list_types', 'list_by_cid', 'list_by_type',
    'list_by_extension', 'findall', 'find',

    # anyconfig.query
    'try_query',

    # anyconfig.validate
    'validate', 'is_valid', 'gen_schema'
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import pathlib
import unittest

from . import constants, utils as TT


RES_DIR = constants.RES_DIR / 'base'
SELF = pathlib.Path(__file__)


class TestCase(unittest.TestCase):

    def test_target_by_parent(self):
        aes = [
            ((), 'base'),
            ((__file__, ), 'base'),
        ]
        for args, exp in aes:
            self.assertEqual(TT.target_by_parent(*args), exp)

    def test_load_from_py(self):
        constants_py_path = SELF.parent / 'constants.py'
        aes = [
            ((constants_py_path, ), constants.DATA),
            ((str(constants_py_path), ), constants.DATA),
            ((constants_py_path, 'RES_DIR'), constants.RES_DIR),
        ]
        for args, exp in aes:
            self.assertEqual(
                TT.load_from_py(*args), exp,
                f'args: {args!r}, exp: {exp!r}'
            )

    def test_load_literal_data_from_py(self):
        py_path = RES_DIR / 'basics' / '20' / '00.py'
        exp = TT.json.load(
            (RES_DIR / 'basics' / '10' / '00.json').open()
        )
        aes = [
            (py_path, exp),
            (str(py_path), exp),
        ]
        for arg, exp in aes:
            self.assertEqual(
                TT.load_literal_data_from_py(arg), exp
            )

    def test_maybe_data_path(self):
        aes = [
            ((SELF.parent, SELF.stem, ), SELF),
            ((pathlib.Path('/not/exist/dir'), 'foo', ), None),
        ]
        for args, exp in aes:
            self.assertEqual(TT.maybe_data_path(*args), exp)

    def test_maybe_data_path_failures(self):
        aes = [
            (SELF.parent, SELF.stem, (SELF.parent.name, ), '.xyz'),
        ]
        for args in aes:
            with self.assertRaises(OSError):
                TT.maybe_data_path(*args)

    def test_load_data(self):
        aes = [
            ((None, ), {}),
            ((None, 1), 1),
            ((RES_DIR / 'basics' / '10' / '00.json', ),
             TT.json.load((RES_DIR / 'basics' / '10' / '00.json').open())
             ),
            ((RES_DIR / 'basics' / '20' / '00.py', ),
             TT.json.load((RES_DIR / 'basics' / '10' / '00.json').open())
             ),
            ((RES_DIR / 'basics' / '30' / '20.txt', ),
             (RES_DIR / 'basics' / '10' / '20.json').read_text()
             ),
        ]
        for args, exp in aes:
            res = TT.load_data(*args)
            self.assertEqual(res, exp, res)

    def test_load_data_failures(self):
        aes = [
            (pathlib.Path('not_exist.xyz'), ),
        ]
        for args in aes:
            with self.assertRaises(ValueError):
                TT.load_data(*args)

    def test_each_data_from_dir(self):
        aes = [
            ((RES_DIR / 'basics' / '10', '*.json', ('c', 'e', 'o', 's')), 3),
            ((RES_DIR / 'basics' / '20', '*.py'), 1),
            ((RES_DIR / 'basics' / '30', '*.txt', ('c', 'e', 'o', 's')), 3),
        ]
        for args, exp in aes:
            res = list(TT.each_data_from_dir(*args))
            self.assertTrue(bool(res))
            self.assertEqual(len(res), exp)

    def test_each_data_from_dir_failures(self):
        with self.assertRaises(ValueError):
            _ = list(TT.each_data_from_dir(SELF))

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import unittest

import anyconfig.query

from . import common


@unittest.skipIf(not anyconfig.query.SUPPORTED,
                 'jmespath lib is not available')
class TestCase(common.TestCase):
    kind = 'query'
    should_exist = ('e', 'q')

    def test_multi_load(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn(
                    tdata.inputs, ac_query=tdata.query, **tdata.opts
                ),
                tdata.exp
            )

    def test_multi_load_with_invalid_query(self):
        for tdata in self.each_data():
            self.assertEqual(
                self.target_fn(tdata.inputs, ac_query='', **tdata.opts),
                self.target_fn(tdata.inputs)
            )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2012 - 2021 Satoru SATOH <satoru.satoh@gmail.com>
# License: MIT
#
# pylint: disable=missing-docstring
import pathlib
import tempfile
import unittest
import warnings

import anyconfig.template

from . import common


@unittest.skipIf(not anyconfig.template.SUPPORTED,
                 'jinja2 template lib is not available')
class TestCase(common.TestCase):
    kind = 'template'
    pattern = '*.j2'

    def test_single_load(self):
        for data in self.each_data():
            self.assertEqual(
                self.target_fn(
                    data.inp_path, ac_context=data.ctx, **data.opts
                ),
                data.exp,
                f'{data.datadir!s}, {data.inp_path!s}'
            )

    def test_single_load_from_invalid_template(self):
        with tempfile.TemporaryDirectory() as tdir:
            wdir = pathlib.Path(tdir)
            inp = wdir / 'test.json'
            inp.write_text('{"a": "{{ a"}')  # broken template string.

            with warnings.catch_warnings(record=True) as warns:
                warnings.simplefilter('always')
                res = self.target_fn(
                    inp, ac_template=True, ac_context=dict(a=1)
                )
                self.assertEqual(res, dict(a='{{ a'))
                self.assertTrue(len(warns) > 0)
                self.assertTrue(issubclass(warns[-1].category, UserWarning))
                self.assertTrue('ailed to compile ' in str(warns[-1].message))

    def test_single_load_intentional_failures(self):
        ng_exp = dict(z=1, zz='zz', zzz=[1, 2, 3], zzzz=dict(z=0))
        for data in self.each_data():
            with self.assertRaises(AssertionError):
                self.assertEqual(
                    self.target_fn(
                        data.inp_path, ac_context=data.ctx, **data.opts
                    ),
                    ng_exp
                )

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""Abstract class provides a list of :class:`anyconfig.models.processor` class
objects, related utility functions and data types.

.. versionchanged:: 0.10.2

   - Split and re-organize the module and add some data types.

.. versionadded:: 0.9.5

   - Add to abstract processors such like Parsers (loaders and dumpers).
"""
from .datatypes import (
    ProcT, ProcClsT, ProcClssT, MaybeProcT
)
from .processors import Processors
from .utils import (
    list_by_x, load_plugins
)

__all__ = [
    'ProcT', 'ProcClsT', 'ProcClssT', 'MaybeProcT',
    'Processors',
    'list_by_x', 'load_plugins',
]

# vim:sw=4:ts=4:et:



#
# Copyright (C) 2021 Satoru SATOH <satoru.satoh@gmail.com>
# SPDX-License-Identifier: MIT
#
r"""ioinfo module to provide functions to create IOInfo objects wrap
pathlib.Path and io objects.

.. versionchanged:: 0.12.0

- Restructure and migrate some utility functions in .utils into this module.

.. versionchanged:: 0.10.1

- simplify inspect_io_obj and make; detect type in make, remove the member
  opener from ioinfo object, etc.

.. versionadded:: 0.9.5

- Add functions to make and process input and output object holding some
  attributes like input and output type (path, stream or pathlib.Path object),
  path, opener, etc.
"""
from .factory import make
from .paths import expand_paths

__all__ = [
    'make', 'expand_paths'
]

# vim:sw=4:ts=4:et:
